@@ -4,28 +4,35 @@
 # With contributions of Marius Lange and Michal Klein.
 # Based on the original MATLAB GPCCA code authored by Bernhard Reuter, Susanna Roeblitz and Marcus Weber,
 # Zuse Institute Berlin, Takustrasse 7, 14195 Berlin
-# --------------------------------------------------
+# ---------------------------------------------------------------------------------------------------------------------
+# The development of pyGPCCA started at the beginning of 2020 in a fork of MSMTools
+# (Copyright (c) 2015, 2014 Computational Molecular Biology Group, Freie Universitaet Berlin (GER);
+# provided under LGPL-3.0 License), since at this time it was planned to integrate GPCCA into it.
+# Due to this, some similarities in structure/AST and code (indicated were evident) between pcca.py
+# https://github.com/markovmodel/msmtools/blob/93126608c6fa9c3197f4fae2f6da93140762b047/msmtools/analysis/dense/pcca.py
+# and _gpcca.py can be found.
+# ---------------------------------------------------------------------------------------------------------------------
 # If you use this code or parts of it, cite the following reference:
-# ------------------------------------------------------------------
-# Reuter, B., Weber, M., Fackeldey, K., Röblitz, S., & Garcia, M. E. (2018).
-# Generalized Markov State Modeling Method for Nonequilibrium Biomolecular Dynamics:
-# Exemplified on Amyloid β Conformational Dynamics Driven by an Oscillating Electric Field.
-# Journal of Chemical Theory and Computation, 14(7), 3579–3594. https://doi.org/10.1021/acs.jctc.8b00079
-# ----------------------------------------------------------------
-# pyGPCCA is free software: you can redistribute it and/or modify
-# it under the terms of the GNU Lesser General Public License as published by
-# the Free Software Foundation, either version 3 of the License, or
-# (at your option) any later version.
+# ---------------------------------------------------------------------------------------------------------------------
+# Bernhard Reuter, Konstantin Fackeldey, and Marcus Weber,
+# Generalized Markov modeling of nonreversible molecular kinetics,
+# The Journal of Chemical Physics, 150(17):174103, 2019.
+# https://doi.org/10.1063/1.5064530
+# ---------------------------------------------------------------------------------------------------------------------
+# pyGPCCA is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser
+# General Public License as published by the Free Software Foundation, either version 3 of the License,
+# or (at your option) any later version.
 #
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
-# GNU General Public License for more details.
+# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
+# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
+# See the GNU General Public License for more details.
 #
-# You should have received a copy of the GNU Lesser General Public License
-# along with this program.  If not, see <http://www.gnu.org/licenses/>.
+# You should have received a copy of the GNU Lesser General Public License along with this program.
+# If not, see <http://www.gnu.org/licenses/>.
+# ---------------------------------------------------------------------------------------------------------------------
+
 __author__ = __maintainer__ = "Bernhard Reuter"
-__email__ = "bernhard.reuter AT uni-tuebingen DOT de"
+__email__ = "bernhard-reuter@gmx.de"
 __copyright__ = "Copyright 2020, Bernhard Reuter"
 __credits__ = [
     "Bernhard Reuter",
@@ -125,7 +132,7 @@ def _gram_schmidt_mod(X: np.ndarray, eta: np.ndarray) -> np.ndarray:
     X[:, 0] = np.sqrt(eta)
     # Raise, if the subspace changed!
     dummy = subspace_angles(X, Xc)
-    if not np.allclose(dummy, 0.0, atol=1e-8, rtol=1e-5):
+    if not np.allclose(dummy, 0.0, atol=1e-7, rtol=1e-5):
         logging.error(Xc)
         logging.error(X)
         raise ValueError(
@@ -146,7 +153,7 @@ def _gram_schmidt_mod(X: np.ndarray, eta: np.ndarray) -> np.ndarray:
 
     # Raise, if the subspace changed!
     dummy = subspace_angles(Q, Xc)
-    if not np.allclose(dummy, 0.0, atol=1e-8, rtol=1e-5):
+    if not np.allclose(dummy, 0.0, atol=1e-7, rtol=1e-5):
         raise ValueError(
             "The subspace of Q derived by eta-orthogonalization doesn't match the "
             f"subspace of the original Q! The subspace angles are: {dummy}. "
@@ -154,7 +161,11 @@ def _gram_schmidt_mod(X: np.ndarray, eta: np.ndarray) -> np.ndarray:
         )
     # Raise, if the (Schur)vectors aren't orthogonal!
     if not np.allclose(Q.conj().T.dot(Q), np.eye(Q.shape[1]), atol=1e-8, rtol=1e-5):
-        raise ValueError("(Schur)vectors do not appear to be orthogonal.")
+        dev = np.max(np.abs(Q.conj().T.dot(Q) - np.eye(Q.shape[1])))
+        raise ValueError(
+            f"(Schur)vectors do not appear to be orthogonal. Largest absolute element-wise deviation in "
+            f"(Q^*TQ - I) is {dev}"
+        )
 
     return Q
 
@@ -222,11 +233,14 @@ def _do_schur(
     if eta.shape[0] != N1:
         raise ValueError("eta vector length doesn't match with the shape of P.")
     if not np.allclose(np.sum(P, 1), 1.0, rtol=1e-6, atol=1e-6):  # previously eps
+        dev = np.max(np.abs(np.sum(P, 1) - 1.0))
         raise ValueError(
-            "Not all rows of P sum up to one (within numerical precision). P must be a row-stochastic matrix."
+            f"Not all rows of P sum up to one. P must be a row-stochastic matrix Largest deviation from row-sums to 1 "
+            f"is {dev}."
         )
     if not np.all(eta > EPS):
-        raise ValueError("Not all elements of eta are > 0 (within numerical precision).")
+        smallest_eta = np.min(eta)
+        raise ValueError(f"Not all elements of eta are > 0. The smallest element is {smallest_eta}")
 
     # Weight the stochastic matrix P by the input (initial) distribution eta.
     if issparse(P):
@@ -272,12 +286,19 @@ def _do_schur(
         )
     # Raise, if the first column X[:,0] of the Schur vector matrix isn't constantly equal 1!
     if not np.allclose(X[:, 0], 1.0, atol=1e-8, rtol=1e-5):
-        raise ValueError("The first column X[:, 0] of the Schur vector matrix isn't constantly equal 1.")
+        dev = np.max(np.abs(X[:, 0] - 1.0))
+        raise ValueError(
+            f"The first column X[:, 0] of the Schur vector matrix isn't constantly equal 1. The largest "
+            f"deviation from one is {dev}."
+        )
 
     # Raise, if the (Schur)vectors aren't D-orthogonal (don't fullfill the orthogonality condition)!
     if not np.allclose(X.T.dot(X * eta[:, None]), np.eye(X.shape[1]), atol=1e-6, rtol=1e-5):
-        logging.error(X.T.dot(X * eta[:, None]))
-        raise ValueError("Schur vectors appear to not be D-orthogonal.")
+        dev = np.max(np.abs(X.T.dot(X * eta[:, None]) - np.eye(X.shape[1])))
+        raise ValueError(
+            f"Schur vectors appear to not be D-orthogonal. The largets deviation of X^T D X from the "
+            f"identity matrix is {dev}"
+        )
 
     # Raise, if X doesn't fullfill the invariant subspace condition!
     dp = np.dot(P, sp.csr_matrix(X) if issparse(P) else X)
@@ -365,8 +386,9 @@ def _indexsearch(X: np.ndarray) -> np.ndarray:
     diffs = np.abs(np.max(X, axis=0) - np.min(X, axis=0))
     if not np.isclose(1.0 + diffs[0], 1.0, rtol=1e-6):
         raise ValueError(
-            "First Schur vector is not constant 1. This indicates that the Schur vectors "
-            "are incorrectly sorted. Cannot search for a simplex structure in the data."
+            f"First Schur vector is not constant 1. This indicates that the Schur vectors "
+            f"are incorrectly sorted. Cannot search for a simplex structure in the data. The largest deviation from 1 "
+            f"is {diffs[0]}."
         )
     if not np.all(diffs[1:] > 1e-6):
         which = np.sum(diffs[1:] <= 1e-6)
@@ -434,7 +456,7 @@ def _objective(alpha: np.ndarray, X: np.ndarray) -> float:
     rot_mat = np.zeros((m, m), dtype=np.float64)
 
     # Sanity checks.
-    if alpha.shape[0] != k ** 2:
+    if alpha.shape[0] != k**2:
         raise ValueError(
             "The shape of alpha doesn't match with the shape of X: "
             f"It is not a ({k}^2,)-vector, but of dimension {alpha.shape}. X is of shape `{X.shape}`."
@@ -492,7 +514,7 @@ def _opt_soft(X: np.ndarray, rot_matrix: np.ndarray) -> Tuple[np.ndarray, np.nda
 
     # Now reshape rot_crop_matrix into a linear vector alpha.
     k = m - 1
-    alpha = np.reshape(rot_crop_matrix, k ** 2)
+    alpha = np.reshape(rot_crop_matrix, k**2)
     # TODO: Implement Gauss Newton Optimization to speed things up esp. for m > 10
     alpha, fopt, _, _, _ = fmin(_objective, alpha, args=(X,), full_output=True, disp=False)
 
@@ -508,13 +530,17 @@ def _opt_soft(X: np.ndarray, rot_matrix: np.ndarray) -> Tuple[np.ndarray, np.nda
 
     # Check for negative elements in chi and handle them.
     if np.any(chi < 0.0):
-        if np.any(chi < -10 * EPS):
-            raise ValueError(f"Some elements of chi are significantly negative: < {-10 * EPS}.")
+        if np.any(chi < -1e4 * EPS):
+            min_el = np.min(chi)
+            raise ValueError(f"Some elements of chi are significantly negative. The minimal element in chi is {min_el}")
         else:
             chi[chi < 0.0] = 0.0
             chi = np.true_divide(1.0, np.sum(chi, axis=1))[:, np.newaxis] * chi
             if not np.allclose(np.sum(chi, axis=1), 1.0, atol=1e-8, rtol=1e-5):
-                raise ValueError("The rows of chi don't sum up to 1.0 after rescaling.")
+                dev = np.max(np.abs(np.sum(chi, axis=1) - 1.0))
+                raise ValueError(
+                    f"The rows of chi don't sum up to 1.0 after rescaling. Maximum deviation from 1 is {dev}"
+                )
 
     return rot_matrix, chi, fopt
 
@@ -603,7 +629,7 @@ def _cluster_by_isa(X: np.ndarray) -> Tuple[np.ndarray, float]:
 @d.dedent
 def _gpcca_core(X: np.ndarray) -> Tuple[np.ndarray, np.ndarray, float]:
     r"""
-    Core of the G-PCCA [Reuter18]_ spectral clustering method with optimized memberships.
+    Core of the G-PCCA spectral clustering method with optimized memberships [Reuter18]_, [Reuter19]_.
 
     Clusters the dominant `m` Schur vectors of a transition matrix.
     This algorithm generates a fuzzy clustering such that the resulting
@@ -646,7 +672,7 @@ def _coarsegrain(P: Union[np.ndarray, spmatrix], eta: np.ndarray, chi: np.ndarra
 
     ..math: P_c = (\chi^T D \chi)^{-1} (\chi^T D P \chi)
 
-    with :math:`D` being a diagonal matrix with `eta` on its diagonal [Reuter18]_.
+    with :math:`D` being a diagonal matrix with `eta` on its diagonal [Reuter18]_, [Reuter19]_.
 
     Parameters
     ----------
@@ -680,14 +706,14 @@ def gpcca_coarsegrain(
     method: str = DEFAULT_SCHUR_METHOD,
 ) -> np.ndarray:
     r"""
-    Coarse-grain the transition matrix `P` into `m` sets using G-PCCA [Reuter18]_.
+    Coarse-grain the transition matrix `P` into `m` sets using G-PCCA [Reuter18]_, [Reuter19]_.
 
     Performs optimized spectral clustering via G-PCCA and coarse-grains `P`
     such that the dominant Perron eigenvalues are preserved using:
 
     .. math:: P_c = (\chi^T D \chi)^{-1} (\chi^T D P \chi)
 
-    with :math:`D` being a diagonal matrix with `eta` on its diagonal [Reuter18]_.
+    with :math:`D` being a diagonal matrix with `eta` on its diagonal [Reuter18]_, [Reuter19]_.
 
     Parameters
     ----------
@@ -706,7 +732,7 @@ def gpcca_coarsegrain(
 
     References
     ----------
-    If you use this code or parts of it, please cite [Reuter18]_.
+    If you use this code or parts of it, please cite [Reuter19]_.
     """
     # Matlab: Pc = pinv(chi'*diag(eta)*chi)*(chi'*diag(eta)*P*chi)
     chi = GPCCA(P, eta=eta, z=z, method=method).optimize(m).memberships
@@ -717,7 +743,7 @@ def gpcca_coarsegrain(
 @d.dedent
 class GPCCA:
     """
-    G-PCCA [Reuter18]_ spectral clustering method with optimized memberships.
+    G-PCCA spectral clustering method with optimized memberships [Reuter18]_, [Reuter19]_.
 
     Clusters the dominant `m` Schur vectors of a transition matrix.
 
@@ -736,7 +762,7 @@ class GPCCA:
 
     References
     ----------
-    If you use this code or parts of it, please cite [Reuter18]_.
+    If you use this code or parts of it, please cite [Reuter19]_.
     """
 
     def __init__(
@@ -878,7 +904,7 @@ class GPCCA:
         m: Union[int, Tuple[int, int], List[int], Dict[str, int]],
     ) -> "GPCCA":
         r"""
-        Full G-PCCA [Reuter18]_ spectral clustering method with optimized memberships.
+        Full G-PCCA spectral clustering method with optimized memberships [Reuter18]_, [Reuter19]_.
 
         It also has the option to optimize the number of clusters
         (macrostates) `m` as well.
@@ -957,6 +983,9 @@ class GPCCA:
         if min(m_list) in [0, 1]:
             raise ValueError(f"There is no point in clustering into `{m}` clusters.")
 
+        # The following code enclosed by >>>... ...<<< originates (with some adjustments) from MSMTools
+        # Copyright (c) 2015, 2014 Computational Molecular Biology Group, Freie Universitaet Berlin (GER).
+        # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
         # test connectivity
         components = connected_sets(self._P)
         n_components = len(components)
@@ -970,6 +999,7 @@ class GPCCA:
             if np.sum(self._P[component, :][:, rest]) == 0:
                 closed_components.append(component)
         n_closed_components = len(closed_components)
+        # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
 
         # Calculate Schur matrix R and Schur vector matrix X, if not adequately given.
         self._do_schur_helper(max(m_list))
@@ -1029,9 +1059,9 @@ class GPCCA:
                     f"Clustering {n} data points into {max(m_list)} clusters is always perfectly crisp. "
                     f"Thus m={max(m_list)} won't be included in the search for the optimal cluster number."
                 )
-                opt_idx = np.argmax(crispness_list[:-1])
+                opt_idx = int(np.argmax(crispness_list[:-1]))
             else:
-                opt_idx = np.argmax(crispness_list)
+                opt_idx = int(np.argmax(crispness_list))
         else:
             raise ValueError("Clustering wasn't successful. Try different cluster numbers.")
         self._m_opt = min(m_list) + opt_idx
@@ -1138,8 +1168,14 @@ class GPCCA:
         -------
         Integer vector of shape `(n,)` containing the macrostate
         each microstate is located in.
+
+        Credits
+        -------
+        The code and docstring of this property origins (with some adjustments) from MSMTools,
+        Copyright (c) 2015, 2014 Computational Molecular Biology Group,
+        Freie Universitaet Berlin (GER).
         """
-        return None if self.memberships is None else np.argmax(self.memberships, axis=1)
+        return None if self.memberships is None else np.argmax(self.memberships, axis=1)  # type: ignore[return-value]
 
     @property
     def macrostate_sets(self) -> Optional[List[np.ndarray]]:
@@ -1155,6 +1191,12 @@ class GPCCA:
         A list of length equal to :attr:`n_m`.
 
         Each element is an array with microstate indexes contained in it.
+
+        Credits
+        -------
+        The code and docstring of this property origins (with some adjustments) from MSMTools,
+        Copyright (c) 2015, 2014 Computational Molecular Biology Group,
+        Freie Universitaet Berlin (GER).
         """
         return (
             None
@@ -1228,7 +1270,7 @@ class GPCCA:
         """
         try:
             return stationary_distribution(self._P)
-        except ValueError as e:
+        except Exception as e:  # noqa: B902
             warnings.warn(f"Stationary distribution couldn't be calculated. Reason: {e}.")
         return None
 
